/***************************************************************************
 * CVSID: $Id: hald.c,v 1.21 2004/10/18 22:52:05 david Exp $
 *
 * hal_addon_usb_csr.c : daemon handling CSR-based wireless mice
 *
 * Copyright (C) 2004 Sergey V. Udaltsov <svu@gnome.org>
 *
 * Licensed under the Academic Free License version 2.0
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **************************************************************************/

#include <config.h>

#include <stdio.h>

#include <usb.h>

#include <glib/gmain.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>

static LibHalContext * hal_context = NULL;

#define CMD_IFACE_PROPERTY "battery.command_interface"
#define BUS_NO_PROPERTY "usb.bus_number"
/* linux!!! */
#define PORT_NO_PROPERTY "usb.linux.device_number"
#define CSR_IS_DUAL_PROPERTY "battery.csr.is_dual"
#define CURRENT_CHARGE_PROPERTY "battery.charge_level.current"

typedef struct _PropertyCacheItem
{
	gboolean cmd_iface_present;
	char* cmd_iface;

	gboolean bus_no_present;
	int bus_no;

	gboolean port_no_present;
	int port_no;

	gboolean csr_is_dual_present;
	gboolean csr_is_dual;

	gboolean current_charge_present;
	int current_charge;
} PropertyCacheItem;

static GHashTable * cached_properties = NULL;

#define TIMEOUT         0x1000

static PropertyCacheItem* 
cached_properties_add(const char * hal_device_udi)
{
	PropertyCacheItem * pci = g_new0 (PropertyCacheItem,1);
	pci->cmd_iface_present = hal_device_property_exists (hal_context,
							     hal_device_udi,
							     CMD_IFACE_PROPERTY);
	if (pci->cmd_iface_present)
	{
		char* ci = hal_device_get_property_string (hal_context,
							   hal_device_udi,
							   CMD_IFACE_PROPERTY);
		if (ci != NULL)
		{
			pci->cmd_iface = g_strdup (ci);
			hal_free_string (ci);
		}
	}

	pci->bus_no_present = hal_device_property_exists (hal_context,
							  hal_device_udi,
							  BUS_NO_PROPERTY);
	if (pci->bus_no_present)
		pci->bus_no = hal_device_get_property_int (hal_context,
							   hal_device_udi,
							   BUS_NO_PROPERTY);


	pci->port_no_present = hal_device_property_exists (hal_context,
							   hal_device_udi,
							   PORT_NO_PROPERTY);
	if (pci->port_no_present)
		pci->port_no = hal_device_get_property_int (hal_context,
							    hal_device_udi,
							    PORT_NO_PROPERTY);


	pci->csr_is_dual_present = hal_device_property_exists (hal_context,
							       hal_device_udi,
							       CSR_IS_DUAL_PROPERTY);
	if (pci->csr_is_dual_present)
		pci->csr_is_dual = hal_device_get_property_bool (hal_context,
								 hal_device_udi,
								 CSR_IS_DUAL_PROPERTY);


	pci->current_charge_present = hal_device_property_exists (hal_context,
								  hal_device_udi,
								  CURRENT_CHARGE_PROPERTY);
	if (pci->current_charge_present)
		pci->current_charge = hal_device_get_property_int (hal_context,
								   hal_device_udi,
								   CURRENT_CHARGE_PROPERTY);

	g_hash_table_insert (cached_properties, g_strdup (hal_device_udi), pci);
	return pci;
}

static void
property_cache_item_free (PropertyCacheItem* pci)
{
	if (pci == NULL)
		return;
	if (pci->cmd_iface != NULL)
 		g_free (pci->cmd_iface);
	g_free (pci);
}

static void
cached_properties_remove(const char * hal_device_udi)
{
	g_hash_table_remove (cached_properties, hal_device_udi);
}

static gboolean
cached_properties_exist (const char * hal_device_udi)
{
	return g_hash_table_lookup (cached_properties, (gconstpointer)hal_device_udi) != NULL;
}

static struct usb_device* 
find_device (const char * hal_device_udi, PropertyCacheItem * pci);

static void
mainloop_integration (LibHalContext *ctx, DBusConnection * dbus_connection)
{
	dbus_connection_setup_with_g_main (dbus_connection, NULL);
}

/* Thanks to lmctl code. I'd LOVE, REALLY LOVE to see some docs though... */
static void 
check_battery (const char * hal_device_udi, PropertyCacheItem * pci)
{
	struct usb_device *current_usb_device;
        usb_dev_handle * handle;
        unsigned char buf[80];
/* Internal CSR registered, I presume - for some reason not addressed directly */
#define P6  (buf[0])
#define P0  (buf[1])
#define P4  (buf[2])
#define P5  (buf[3])
#define P8  (buf[4])
#define P9  (buf[5])
#define PB0 (buf[6])
#define PB1 (buf[7])
        unsigned int  addr;
	int is_dual = 0;

	if (pci == NULL)
		return;

	printf ("CSR device: [%s]\n", hal_device_udi);

	is_dual = pci->csr_is_dual;

	/* Which of subdevices to address */
	printf ("Is dual: %d\n", is_dual);
        addr = is_dual? 1<<8 : 0;

	current_usb_device = find_device (hal_device_udi, pci);
	if (current_usb_device == NULL)
	{
		fprintf (stderr, "Device %s not found\n", hal_device_udi);
		return;
	}

        handle = usb_open (current_usb_device);
	if (handle == NULL)
	{
		perror ("Could not open usb device\n");
		return;
	}

	if (!usb_control_msg (handle, 0xc0, 0x09, 0x03|addr, 0x00|addr,
			      buf, 8, TIMEOUT) != 8)
	{
		if ((P0 == 0x3b) && (P4 == 0))
		{
			printf ("Receiver busy, trying again later\n");
		} else
		{
			int current_charge = P5 & 0x07;

			printf ("Charge level: %d->%d\n", 
				pci->current_charge, current_charge);
			if (current_charge != pci->current_charge)
			{
				pci->current_charge = current_charge;
		    		hal_device_set_property_int (hal_context, 
							     hal_device_udi, 
							     CURRENT_CHARGE_PROPERTY,
							     current_charge);
			}
		}
	} else
	{
		perror("Writing to USB device");
        }

	usb_close (handle);
}

/* TODO: Is it linux-specific way to find the device? */
static struct usb_device* 
find_device (const char * hal_device_udi, PropertyCacheItem * pci)
{
	struct usb_bus* current_usb_bus;
	char LUdirname[5];
	char LUfname[5];

	if (!(pci->bus_no_present && pci->port_no_present))
	{
		/* no sysfs path */
		fprintf (stderr, "No hal bus number and/or port number\n");
		return NULL;
	}
	snprintf (LUdirname, sizeof (LUdirname), "%03d", pci->bus_no);
	snprintf (LUfname, sizeof (LUfname), "%03d",pci->port_no);
	printf( "Looking for: [%s][%s]\n", LUdirname, LUfname);

	for (current_usb_bus = usb_busses; 
	     current_usb_bus != NULL; 
	     current_usb_bus = current_usb_bus->next)
	{
        	struct usb_device *current_usb_device;
		/* printf ("Checking bus: [%s]\n", current_usb_bus->dirname); */
		if (g_strcasecmp (LUdirname, current_usb_bus->dirname))
			continue;

        	for (current_usb_device = current_usb_bus->devices; 
		     current_usb_device != NULL; 
		     current_usb_device = current_usb_device->next) 
		{
			/* printf ("Checking port: [%s]\n", current_usb_device->filename); */
			if (g_strcasecmp (LUfname, current_usb_device->filename))
				continue;
			printf( "Matched device: [%s][%s][%04X:%04X]\n", 
				current_usb_bus->dirname, 
				current_usb_device->filename,
				current_usb_device->descriptor.idVendor,
				current_usb_device->descriptor.idProduct);
			return current_usb_device;
		}
	}
	return NULL;
}

static gboolean
check_all_batteries (gpointer data)
{
	printf ("** Check batteries\n");

	/* TODO: make it configurable (not to rescan every time) */
	usb_find_busses();
	usb_find_devices();

	g_hash_table_foreach (cached_properties, (GHFunc)check_battery, NULL);

	return TRUE;
}

static void
device_added (LibHalContext *ctx, const char *hal_device_udi)
{
	if (hal_device_property_exists (hal_context, 
					hal_device_udi, 
					CMD_IFACE_PROPERTY))
	{
		char* ci = hal_device_get_property_string (hal_context, 
							   hal_device_udi, 
							   CMD_IFACE_PROPERTY);
		if (ci != NULL)
		{
			if (!g_strcasecmp ("csr", ci))
			{
				printf ("** CSR device added: %s\n", hal_device_udi);
				cached_properties_add (hal_device_udi);
			}
			hal_free_string (ci);
		}
	}
}

static void
device_removed (LibHalContext *ctx, const char *hal_device_udi)
{
	if (cached_properties_exist (hal_device_udi))
	{
		printf ("** CSR device removed: %s\n", hal_device_udi);
		cached_properties_remove (hal_device_udi);
	}
}

static void 
property_modified (LibHalContext *ctx,
		   const char *hal_device_udi,
		   const char *key,
		   dbus_bool_t is_removed,
		   dbus_bool_t is_added)
{
	/* "Key" property modified */
	if (!g_ascii_strcasecmp (key, CMD_IFACE_PROPERTY))
	{
		if (is_removed)
		{
			printf ("** Property %s removed: %s\n", key, hal_device_udi);
			cached_properties_remove (hal_device_udi);
		} else
		{
			char* ci = hal_device_get_property_string (hal_context, 
								   hal_device_udi, 
								   CMD_IFACE_PROPERTY);
			if (ci != NULL)
			{
				if (!g_strcasecmp ("csr", ci))
				{
					printf ("** Property %s added/changed: %s\n", 
						key, hal_device_udi);
					cached_properties_add (hal_device_udi);
				}
				hal_free_string (ci);
			}
		}
		
	}

	/* "Secondary" property modified */
	if (cached_properties_exist (hal_device_udi))
	{
		if (!(g_ascii_strcasecmp(key, BUS_NO_PROPERTY) &&
		      g_ascii_strcasecmp(key, PORT_NO_PROPERTY) &&
	    	      g_ascii_strcasecmp(key, CSR_IS_DUAL_PROPERTY)))
		{
			printf ("** Property %s added/changed: %s\n", 
				key, hal_device_udi);
			cached_properties_remove (hal_device_udi);
			cached_properties_add (hal_device_udi);
		}
	}
}

static void
initial_fillup (void)
{
	int num_devices, i;
	printf ("** Initial fillup\n");
	char** all_devices = hal_get_all_devices (hal_context, &num_devices);
	if (all_devices != NULL && num_devices != 0)
	{
		char** current_device = all_devices;

		printf ("Check all CSR devices:\n");
		for(i=num_devices;--i>=0; current_device++)
		{
			if (hal_device_property_exists (hal_context, 
							*current_device, 
							CMD_IFACE_PROPERTY))
			{
				char* ci = hal_device_get_property_string (hal_context, 
									   *current_device, 
									   CMD_IFACE_PROPERTY);
				if (ci != NULL)
				{
					if (!g_strcasecmp ("csr", ci))
					{
						printf ("CSR device found: [%s]\n", *current_device);
						cached_properties_add (*current_device);
					}
					hal_free_string (ci);
				}
			}
		}
	}
	printf ("** Initial fillup done\n");
}

int
main (int argc, char *argv[])
{
	/* TODO: make it configurable*/
	long check_interval = 10L;

	LibHalFunctions functions = {
		mainloop_integration,
		device_added,
		device_removed,
		NULL, NULL,
		property_modified
	};

	GMainLoop* loop;

	cached_properties = g_hash_table_new_full (g_str_hash, g_str_equal, 
			(GDestroyNotify)g_free, (GDestroyNotify)property_cache_item_free);

	hal_context = hal_initialize (&functions, TRUE);
	initial_fillup ();

	usb_init();
	
	printf ("Addon started\n");

	loop = g_main_loop_new (NULL, FALSE);

	g_timeout_add (1000L * check_interval, check_all_batteries, NULL);
	check_all_batteries (NULL);

	g_main_loop_run (loop);

	hal_shutdown (hal_context);
	printf ("Addon exits normally\n");
	return 0;
}
